"""
discord_ui.slash.ext
~~~~~~~~~~~~~~~~~~~~~~
An extension module to the libary that has some useful decorators and functions 
for application-commands.

.. code-block::

    from discord_ui import ext

"""

from .builder import *
from .command_decorators import *